/*
  EPROM Programmer + SD card read/write

  Arduino Mega 2560 plus USB-Tiny programmer


  Mega2650
  USB-ASP

  Now with iterative programming pulses

 ** SD card attached to SPI bus as follows:
 ** MOSI - pin 50
 ** MISO - pin 51
 ** CLK  - pin 52
 ** CS   - pin 53

   looking to program the 8kb EPROMs so looking at 2764's

   Port  MegaPin Vpp   1         28 Vcc      Port Mega PIN
    K    12      A12   2         27 !PGM     K     15
    A    29      A7    3         26 NC       Vcc
    A    28      A6    4         25 A8       K     8
    A    27      A5    5         24 A9       K     9
    A    26      A4    6         23 A11      K     11
    A    25      A3    7         22 !OE      K     13
    A    26      A2    8         21 A10      K     10
    A    25      A1    9         20 !CE      K     14
    A    24      A0   10         19 D7       C     37
    C    30      D0   11         18 D6       C     36
    C    31      D1   12         17 D5       C     35
    C    32      D2   13         16 D4       C     34
    0V                14         15 D3       C     33




*/

#include <SD.h>
File root;

File myFile;
char* epromtype[7] = {"Undefined", "EPROM 2716", "EPROM2732", "EPROM2532", "ROM 4K", "ROM 8K", "EPROM2764"};
int   epromnos = 0; //set to index the EPROM type
char  nomifile[14] = "No File.??."; //target name for file activity
char  inBuff[14] =   "           "; //Holds any keyboard inputs.
// Pin 13 has an LED connected on most Arduino boards.
int   led     = 13;       //User feedback
int   address = 0;        //Dummy address counter
byte  lf      = 0;        //Counter for line feeds
int   addMax;             //eg a 8192 1FFF, or 4096 0FFF ROM size
char  ascii[17];          // stores ASCII version if possible
byte  sdBuff[256];        // Stores binary bytes to be written to SD card
boolean fullHex = true;  // True give full dump of the addresses, data and ASCII etc
int   eprom;              //set Eprom Type, 1=2716, 2=2732, 3=2532, 4="4k ROM", 5="8k ROM", "6=8k EPROM"
int   function;           //choose function: 0.List 1.Dump, 2.Read  3. Erased? 4.Burn 5.Verify

// **********the setup routine runs once when you press reset******************
void setup() {
  DDRC  = B00000000;  // EPROM Data Pins D0-D7      Mega Pins 37-30     Port C  0x11111111
  DDRA  = B11111111;  // EPROM Address Pins A0-A7   Mega Pins 22-29     Port A  0x11111111
  DDRK  = B11111111;  // EPROM Address Pins A8-A12  Mega Pins A8-A12    Port K  0x00011111
  //  plus Chip Enables        Mega Pins A14,A13   !CE !OE (!G)        Port K  0x11100000
  //  plus Programming pulse   Mega Pin  A15       !PP
  //  plus Programming Voltage control   D7        Vpp
  PP(1);   //Programming Pulse
  OE(1);   //Disable NOT outputs enable Port K  0x01000000 Mega pin 13
  CE(1);   //Disable NOT chip enable    Port K  0x00100000 Mega Pin 14
  Vpp(0);  //Disable programming Voltage PinD7
  // Open serial communications and wait for port to open:
  Serial.begin(115200);
  while (!Serial) {
  }
  pinMode(7, OUTPUT); // Control Programming Voltage with Pin 7
  digitalWrite(7, 0); // Turn off the Programming voltage immediately
  Serial.println("\n\n\n\n\n\n********************************");
  Serial.println("*      EPROM Programming       *");
  Serial.println("* 5V, 24/28Pin, 2716,2532,2764 *");
  Serial.println("********************************");
  Serial.print("Initializing SD Filing System...");
  pinMode(53, OUTPUT);//make sure CS is an output
  if (!SD.begin(53)) {
    Serial.println('SDFS Initialization failed!\nPls Check card is inserted');
    while (true) {
      //stop here, go no further
    }
  }
  Serial.println("OK!");
  Serial.println("Current files:..");
  root = SD.open("/");

  //Initialize the digital debug LED pin as an output.
  pinMode(led, OUTPUT);
  lf = 0; //LineFeed counter for formatting
  ascii[17] = 0; //string terminator for 16 ASCII dumps

  root = SD.open("/"); //Create active SD object
  printDirectory(root, 0); //Dump the Root folder

  chooseEPROM(); //Set up for a given EPROM
}

// **********this loop routine runs forever***************
void loop() {
  DDRC  = B00000000;        //Hi impedance inputs Bidirectional Data Bus
  DDRA  = B11111111;        //All outputs Lo Address Mega Port is A0-A7
  DDRK  = B11111111;        //All outputs Hi Address Mega Port is K0-K4, K5 K6 K7 reserved for programming
  CE(1);  //Disable NOT outputs enable    K6, ie A14
  OE(1);  //Disable NOT chip enable       K5, ie A13
  PP(1);  //Disable NOT programming Pulse K7, ie A15  //  plus Chip Enables        Mega Pins A14,A13   !CE !OE (!G)        Port K  0x11100000
  //  plus Programming pulse   Mega Pin  A15       !PP
  //  plus Programming Voltage control   D7        Vpp
  Vpp(0); //Turn off the Programming Voltage

  // this will clear the screen in PUTTY but not Arduino Terminal
  //Serial.write(27);       // ESC command
  //Serial.print("[2J");    // clear screen command
  //Serial.write(27);
  //Serial.print("[H");     // cursor to home command


  Serial.println("\n\n..................................");
  Serial.println("EPROM/ROM Reader/Copier/Programmer");
  Serial.println("''''''''''''''''''''''''''''''''''");
  Serial.print("Nominated filename: ");
  Serial.print(nomifile);// Can be altered in function selection
  existingfile(); //checks existence of file
  Serial.print(epromtype[epromnos]);
  Serial.print(" Memory size 0x");
  Serial.print(addMax, HEX);
  Serial.println(" bytes");
  //Choose the function to perform
  Serial.println(" S  Choose Size of EPROM/ROM");
  Serial.println(" L  List SD Files");
  Serial.println(" F  Choose a File name");
  Serial.println(" I  Inspect an Image file in HEX");
  Serial.println(" E  Check for Erasure");
  Serial.println(" B  Burn File to a ROM");
  Serial.println(" C  Compare a ROM against a File");
  Serial.println(" D  Dump ROM in HEX and ASCII");
  Serial.println(" G  Get a ROM into a File");
  Serial.println(" T  Test OE/EN/PP Pins");
  //delayPP();
  Serial.print("Which Function? ");
  function = getChar(); //get first character from command line
  Serial.println();
  Serial.println();
  if (function == 'S') {
    chooseEPROM();        //Choose the size of EPROM or ROM
  }
  if (function == 'G') {
    writeROM2file(addMax);//Transfer ROM image into a file
  }
  if (function == 'E') {
    erasedROM(addMax);    //Check for all 0xFF's in ROM
  }
  if (function == 'L') {
    dumpSD();            //List the Files on the SD
  }
  if (function == 'F') { //List all files and set the active Filename
    dumpSD();
    setFilename();
  }
  if (function == 'I') { //Dump Active File in Hex
    hexSD(addMax);
  }
  if (function == 'B') {
    burnROM(addMax);     //Burn Image file to ROM
  }
  if (function == 'C') {
    compROM(addMax);     //Image File compared to the ROM onboard
  }
  if (function == 'D') {
    dumpROM(addMax);     //Dump the contents in HEX and ASCII
  }
  if (function == 'T') {
    progTest();          //Flash the control lines CE OE and VP (Vpp) in turn.
  }
}

//Subroutines.....

void hexSD(int epromSize) {
  boolean fullHex = true;
  Serial.print("Reading from ");
  Serial.println(nomifile);
  myFile = SD.open(nomifile, FILE_READ);
  if (myFile) {
    myFile.seek(0);//Send back to the start
    Serial.print("\nDumping a file to Hexadecimal: ");
    Serial.println(nomifile);
    Serial.println("EPROM   : 00 01 02 03 04 05 06 07 08 09 0A 0B 0C 0D 0E 0F");
    Serial.println("---- -- + -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- --");
    Serial.print("Add ");
    hex4Print(0);
    Serial.print(": ");
    //Main loop to read the data from the EPROM
    for ( address = 0; address < epromSize; address++) {
      if (lf == 16) {    // Print the leading name of the line
        if (fullHex) {
          Serial.print(ascii); //Add this to the end of the line
          Serial.println();
          Serial.print("Add "); //Set up start of next line
          hex4Print(address);
          Serial.print(": ");
        }
        else {
          Serial.println();
        }
        lf = 0;
      }
      DDRC = B00000000;               //Data direction Input 8 bits D7-D0
      PORTA = address & 0xFF;         //Lower 8 bits into A port  A7-A0
      PORTK = (address >> 8) & 0x1F;     //Upper 8 bits into K port PP,CE,OE,A12-A8
      PP(1);//PP Programming Pulse Off
      CE(0); //set read enables
      OE(0);
      delay(1);
      byte mybyte = myFile.read(); //read sequential byte from the file
      hex2Print(mybyte);
      Serial.print(" ");
      //now make a list of any printable ASCII equivalents
      if ((mybyte > 31) && (mybyte < 127)) {
        ascii[lf] = mybyte;
      }
      else {
        ascii[lf] = 46; //'.', unprintable ASCII
      }
      lf++;
    }
    if (fullHex) {
      Serial.print(ascii);
      Serial.println();
    }
    CE(1);
    OE(1);
    PP(1);
    Serial.println("Hex/ASCII Dump of file completed");
    Serial.println();

  }
  myFile.close();
}


void delayPP() {
  for (int i = 0; i < 5; i++) {
    PP(0);
    delay(1000);
    PP(1);
    delay(1000);
  }
}

char  chooseEPROM() { //Choose an EPROM Size
  Serial.println("'''''''''''''''''''");
  Serial.println("1. 2716    2k EPROM");
  Serial.println("2. 2732    4k EPROM");
  Serial.println("3. 2532    4k EPROM");
  Serial.println("4. MOS ROM 4k   ROM");
  Serial.println("5. MOS ROM 8k   ROM");
  Serial.println("6. 2764    8k EPROM");

  Serial.print("Which size EEPROM To use? ");
  eprom = getChar(); //get first character from command line
  Serial.write(eprom);
  Serial.println();
  Serial.println();
  //Default is 2732M
  addMax = 0x1000; // bytes
  epromnos = 2;
  //Now select EPROM/ROM
  if (eprom == '1') {
    addMax = 0x800; //2048 bytes
    epromnos = 1;
  }
  if (eprom == '2') {
    addMax = 0x1000; //4096 bytes
    epromnos = 2;
  }
  if (eprom == '3') {
    addMax = 0x1000; //4096 bytes
    epromnos = 3;
  }
  if (eprom == '4') {
    addMax = 0x1000; //4096 bytes
    epromnos = 4;
  }
  if (eprom == '5') {
    addMax = 0x2000; //8192 bytes
    epromnos = 5; //Actually a ROM most likely
  }
  if (eprom == '6') {
    addMax = 0x2000; //8192 bytes
    epromnos = 6; //8kb EPROM
  }
}

void dumpSD() {
  //List all files in the card with date and size
  Serial.println("Contents of the SD Card");
  root = SD.open("/");
  printDirectory(root, 0);
}


//Dump ROM to Screen
void dumpROM(int epromSize) {
  DDRC = B00000000;  //Data direction Input 8 bits
  DDRA = B11111111;  //Data direction Output Lower 8 bits
  DDRK = B11111111;  //Data direction Output Upper 4 bits + CTRL
  lf = 0; //reset the ASCII character index
  if (fullHex) {
    Serial.println("EPROM   : 00 01 02 03 04 05 06 07 08 09 0A 0B 0C 0D 0E 0F");
    Serial.println("---- -- + -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- --");
    Serial.print("Add ");
    hex4Print(0);
    Serial.print(": ");
  }
  //Main loop to read the data from the EPROM
  for ( address = 0; address < epromSize; address++) {
    if (lf == 16) {    // Print the leading name of the line
      if (fullHex) {
        Serial.print(ascii); //Add this to the end of the line
        Serial.println();
        Serial.print("Add "); //Set up start of next line
        hex4Print(address);
        Serial.print(": ");
      }
      else {
        Serial.println();
      }
      lf = 0;
    }
    DDRC = B00000000;               //Data direction Input 8 bits D7-D0
    PORTA = address & 0xFF;         //Lower 8 bits into A port  A7-A0
    PORTK = (address >> 8) & 0x1F;  //Upper 8 bits into K port MSB (PP,CE,OE,A12-A8) LSB
    PP(1);//PP Programming Pulse Off
    CE(0); //set read enables
    OE(0);
    delay(1);
    byte outPins = PINC;            //Read EPROM Data
    hex2Print(outPins);
    Serial.print(" ");
    //now make a list of any printable ASCII equivalents
    if ((outPins > 31) && (outPins < 127)) {
      ascii[lf] = outPins;
    }
    else {
      ascii[lf] = 46; //'.', unprintable ASCII
    }
    lf++;
  }
  if (fullHex) {
    Serial.print(ascii);
    Serial.println();
  }
  CE(1);
  OE(1);
  PP(1);
  Serial.println("Dump completed");
  Serial.println();
}

//Read a byte, and write it into to a file image of the ROM
void writeROM2file(int epromSize) {
  Vpp(0); //Vpp OFF for safety, not necessary with switch
  address = 0; //Main address counter through the binary image
  Serial.print("\n\nWriting memory data to Nominated filename: ");
  Serial.print(nomifile);// Can be altered in function selection
  existingfile(); //checks existence of file
  Serial.print("\nProceed???? Y-N");
  if (getChar() == 'Y') {
    if (SD.exists(nomifile)) {
      SD.remove(nomifile); //Begin a new version
    }
    myFile = SD.open(nomifile, FILE_WRITE);
    // if the file opened okay, write to it:
    if (myFile) {
      myFile.seek(0);//Send back to the start
      Serial.print("\nWriting to Binary ROM to File: ");
      Serial.print(nomifile);
      Serial.print("\n >");
      DDRC = B00000000;  //Data direction Input 8 bits
      DDRA = B11111111;  //Data direction Output Lower 8 Address bits
      DDRK = B11111111;  //Data direction Output Upper 4 Address bits +4CTRL
      PP(1); //Disable programming pulse
      CE(0); //Enable the ROM for a Read
      OE(0); //Enable Output for a Read
      int sdPoint = 0;
      for (int address = 0; address < epromSize; address++) {
        PORTA = address & 0xFF;
        PORTK = ((address >> 8) & 0x0F) | (PORTK & 0xF0); //conserve functions pins
        delay(1);//allow the pins to settle
        byte myByte = PINC;  //Get EPROM Data byte
        //myFile.write(myByte);
        sdBuff[sdPoint] = myByte;
        sdPoint++; //Increase index after transfer, so it is +1
        if (sdPoint == 256) {
          //Write 0-255 bytes to file, ie count of 256 bytes
          myFile.write(sdBuff, 256);//The Count of the bytes
          sdPoint = 0;
          Serial.print("*");  //progress indicator
        }
      }
      // close the file
      myFile.flush();
      myFile.close();
      Serial.println("<  \nDone!");
      Serial.println();
    }
    else {
      // if the file didn't open, print an error:
      Serial.print("\nError opening :");
      Serial.println(nomifile);
    }
  }
  else
  {
    Serial.println("\n Writing cancelled....");
  }
  Serial.println();
  CE(1);
  OE(1);
}


void erasedROM(int epromSize) {
  DDRC = B00000000;  //Read mode
  DDRA = B11111111;  //Address Lower 8 bits
  DDRK = B11111111;  //Address Upper 4 bits +4CTRL

  lf = 0;
  int errors = 0;
  for (int address = 0; address < epromSize; address++) {
    //Write the byte from the EPROM to a file
    PORTA = address & 0xFF; //Lower 8 bits Address
    PORTK = ((address >> 8) & 0x0F); //Upper 4 bits Address
    Vpp(0);// Programming Voltage OFF
    CE(0); //low for read
    OE(0); //low for read
    PP(1); //Programming Pulse off
    delay(1);
    byte myByte = PINC; //Read EPROM Data
    if (myByte != 0xFF) {
      Serial.print("X"); //Not 0xFF found
      errors++;
    }
    else {
      Serial.print("."); //An 0xFF found
    }
    if (((address + 1) % 64) == 0) {
      Serial.println();  //Format the printout
    }
  }
  PP(1);//PP=1=off, ie active low
  CE(1); //Disable ROM
  OE(1); //Disable Outputs
  Serial.println();
  Serial.print(errors, DEC);
  Serial.println(" non-Erased Bytes found!");
  Serial.println();
}

//Write bytes from a File to EPROM
//Maximum EEPROM Size is 80172
//Maybe slower but more comprehensive
void burnROM(int epromSize) {
  // Open the file to write to EPROM:
  Serial.print("Reading from ");
  Serial.println(nomifile);
  myFile = SD.open(nomifile, FILE_READ);
  //EEPROM file Image ready
  int progPulse = 5; //Program for 5ms at a time
  int maxFails = 10; //If more than 10 bytes can't be programmed, abort session
  int maxOverburn = 0; //just checking on EPROM Health
  int minOverburn = 100; //just checking on EPROM Health
  DDRC = B11111111;  //Data direction Output 8 bits Data EPROM D7-D0
  DDRA = B11111111;  //Data direction Output 8 bits, Lower EPROM Address A7-A0
  DDRK = B11111111;  //Data direction Output 8 bits, Upper EPROM Address A12-A8 + Programming Control(5=OE,6=CE,7=PP )
  /*
     Check this if different EPROMS than 2764's are ver used
    if (eprom == 1) {
    PORTK = PORTK | 0x40;  //CE (K6) is hi, so 2532 & 2732 EPROM and programming disabled
    maxFails = 5; //How many times it tries to program the byte
    progPulse = 50; //how long to pulse program pin ms
    }
    else {
    maxFails = 40; //How many times it tries to program the byte
    progPulse = 2; //how long to pulse program pin ms
    }
  */
  int bytes = 0;  //counts bytes for use in progress indicators
  Serial.println("Now Turn on Programming Voltage");
  Serial.print("Ready to Proceed (Y/N)? ");
  Vpp(1);//Enable Programming voltage, or use manual switch
  char response = getChar();
  boolean failure = false; //records overall success or failure of the burn
  Serial.println();
  if (response == 'Y') {
    lf = 0; //format the lines printed
    int burnFails = 0;     //keep track of byte failures and limit max nos of failed bytes
    Serial.print("Now programming an EPROM of 0x");
    Serial.print(epromSize, HEX);
    Serial.println(" bytes");
    //Main byte counter through the binary image
    for (int address = 0; address < epromSize; address++) {
      int success = 0;     //when a byte has burnt once successfully, calculate overburn cycles required
      int overBurn = 20;   //Maximum number of initial tries to get a successful burn
      //on the a successful burn, increase this to 1.5 times ensure 'success'
      int tries = 0;       //reset to zero each byte, and tracks the number of interative loops within each byte burn
      int fails = 0;       //reset for each address, counts number of failed bytes
      byte currentbyte = myFile.read(); //read sequential byte from the file
      /*
        //Enable for monitoring bytes
        hex2Print(currentbyte);
        if ((address % 16) == 0) {
        Serial.println(); //chuck in a bit of formatting
        }
      */
      //Skip programming pulse if the current byte is all ones 0xFF
      if (currentbyte != B11111111) {
        //This do loop controls the nos of program pulses to achieve a burn, or hit a limit?
        do {
          //Set up for programming a byte and act on whether it works or not
          DDRC = B11111111;  //Data direction Output 8 bits Data(D0-D7)
          DDRA = B11111111;  //Data direction Output 8 bits Lower Address (A0-A7)
          DDRK = B11111111;  //Data direction Output 5 bits Upper Address (A8-A12) + 3 bits Programming Control (OE,CE,PP)
          PORTC = currentbyte; //Data to be programmed
          PORTA = address & 0xFF; //Set lower 8 address bits
          PORTK = ((address & 0x1F00) >> 8) | (PORTK & 0xE0); //Set upper 5 address bits (A8-A12) into byte, OE 1, CE 1, PP 0
          Vpp(1); //Programming voltage on, also drives LED
          CE(0); //chip selected for burn
          OE(1); // Output disabled, in high impedance
          delay(1);//Allow Vpp to settle, capacitor to charge
          //get ready to program
          DDRC  =  B11111111; //Port C to Output mode, present data
          PORTC = currentbyte;//Set template Data Byte into Port C
          PP(0);//Start programming pulse, enable
          delay(progPulse); //Using repeated short pulses per byte to reduce potential damage
          PP(1);//Finish programming pulse
          tries++; //keep a track of the number of times required to achieve good programmed data
          //Prepare for Verify
          delay(1);                             //Allow Vpp to settle to OFF
          PORTC =  B00000000;                   //ReSet Cleared Data Byte into Port C, so not reading old C data
          DDRC  =  B00000000;                   //Port C to Input mode
          OE(0);                                //Enable Chip Output for reading
          delay(1);                             //Allow data to settle
          byte checkByte = PINC;                //Read back fresh copy of (hopefully) written byte
          OE(1);                                //Disable Chip Output to continue programming
          //try to get a positive burn but don't go too far, but do it 1+2 times over when successful
          if (checkByte != currentbyte) {
            fails++;                            //Attempts for a given byte
            //Serial.print("A");                //Can be used to monitor programming cycles
          }
          else {
            if (success == 0) {                 //First successful burn, set this only once
              //Is this enough overburn? Some literature suggested 32times overburn????
              overBurn = tries * 2;             //Double of first burn success, ie overburn by 100%
              fails = 0;                        //Fails at beginning are cancelled
            }
            success = 1;                        //Toggle value Successful byte burn
            //Serial.print("S");                //Can be used to monitor programming cycles
          }//end counting successful burns
        }
        //Land here if programming 0xFF

        //once we have success with so many tries we make "overBurn twice" tries
        //and continue until we have "overburned twice" tries.
        //If fails exceeds maxfails something is seriously wrong
        //while byte attempts<maxAttempts and if success, then double it
        while (tries <= overBurn);
        //End of trying to program a Byte
      }
      if (fails > 0) {
        //The byte has failed to program within 25 pulses, possibly bad EPROM
        burnFails++; //Count total nos of bytes failed to programm at all.
        if (burnFails > 20) {
          address = epromSize; //Forces an exit to the loop
          failure = true; //Don't do any more burning
        } //end of checking the limit to nos of failed burns
        //end of checking for failed burns
      }
      if (minOverburn > overBurn) {
        minOverburn = overBurn;
      }
      if (maxOverburn < overBurn) {
        maxOverburn = overBurn;
      }

      //Make a progress printout to show how burn is going
      bytes++;
      //Now update progress indicators in 1/8th jumps
      if ((bytes % (epromSize / 8)) == 0) {
        float prog = float(address) * 100 / epromSize;
        Serial.print("\nProgress ");
        Serial.print(prog, 1);
        Serial.print("%, ");
        Serial.print(burnFails, DEC);
        Serial.println(" Failed");
        bytes = 0;
      } //end of progress report
    } //end of doing all address loops
    Vpp(0); //Remove programming voltage
    Serial.println();
    Serial.println("Burning Attempt Finished");
    Serial.print("minOverburn=");
    Serial.print(minOverburn, DEC);
    Serial.print(", maxOverburn=");
    Serial.println(maxOverburn, DEC);
    Serial.println("Programming Voltage removed");
    if (failure) {
      Serial.println("Programming Aborted! Too many errors!\n\rEPROM probably faulty.");
    }
    else {
      Serial.println("Programming Successful!");
    }
  }//end of burn cycle
  else {
    Serial.println("OK, Burning Cancelled");
    Vpp(0);
  }
  myFile.close();
  Serial.print("Then Enter any key....");
  response = wait_for_char();
  Serial.write(response);
  Serial.println();
  Serial.println("Done.");
  PP(1); //PP=off, ie active low
  CE(1); //Disable ROM
  OE(1); //Disable Outputs
}

//Use the SD file to check agains a programmed ROM
void compROM(int epromSize) {
  myFile = SD.open(nomifile, FILE_READ);
  // if the File opened okay, Read from it:
  if (myFile) {
    myFile.seek(0);//Send back to the start
    Serial.print("Comparing Binary ROM to File: ");
    Serial.println(nomifile);
    Serial.print(" >");
    DDRA = B11111111;  //Data direction Output Lower 8 bits
    DDRK = B11111111;  //Data direction Output Upper 8 bits
    DDRC = B00000000;  //Data direction Input 8 bits, reading from EPROM
    CE(0); //Chip Enabled
    OE(0); //Output Enabled
    PP(1); //Programming Pulse disabled
    Vpp(0); //Programming Voltage off
    int errors = 0;
    for (int address = 0; address < epromSize; address++) {
      //Read byte from the EPROM and compare to file
      if ((address % 256) == 0) {
        Serial.print("*");  //progress indicator
      }
      byte fiByte = myFile.read(); //Read byte from file
      PORTA = address & 0xFF;   //Update Lower 8 address bits
      PORTK = ((address >> 8) | 0xE0); //Retain upper 3 Control bits
      PORTC =  B00000000; //ReSet 'incorrect' Data Byte into Port C, so not reading old C data
      delay(2); //let the addresses settle
      OE(0);
      CE(0);
      byte myByte = PINC; //read byte from EPROM
      if (fiByte != myByte) {
        errors++; //EPROM does not match File byte
        hex4Print(address);
        Serial.print(": File>");
        hex2Print(fiByte); //from file
        Serial.print(" != ROM>");
        hex2Print(myByte); //from ROM
        Serial.println();
      }
    }
    // close the file
    myFile.flush();
    myFile.close();
    Serial.println("<");
    //Print Summary
    Serial.print(errors, DEC);
    Serial.println(" Errors");
  }
  else {
    // if the file didn't open, print an error:
    Serial.print("Error opening ");
    Serial.println(nomifile);
  }
  Serial.println();
  PP(1);//PP=1=off, ie active low
  CE(1); //Disable ROM
  OE(1); //Disable Outputs
}

//Toggle the function pins off and on to test them
//Ideally have LEDs wired to these ports to test them
//This is used as a debug during construction only.
void progTest() {
  //Do not enable if you have an EPROM in the sockets!
  DDRK = B11111111; //Just make an output
  PORTK = 0; //all off
  delay(1000);

  CE(0); //Normally at 1
  delay(1000);
  CE(1);
  Serial.print("CE(A14), "); //Green
  delay(1000);

  OE(0); //Normally at 1
  delay(1000);
  OE(1);
  Serial.print("OE(A13), "); //Yellow
  delay(1000);

  PP(0); //Normally at 1, disabled
  delay(1000);
  PP(1);
  Serial.print("PP (A15), "); //Red
  delay(1000);

  Vpp(1); //Normally at 0  //Orange
  Serial.println("Vpp (D7)");
  delay(1000);
  Vpp(0);
  delay(1000);
}


//General Subroutines

//Format a 8bit data into two hexadecimal numbers
void hex2Print(byte temp) {
  if (temp < 0x10) {
    Serial.print("0");
  }
  Serial.print(temp, HEX);
}

//Format a 16bit address into 4 hexadecimal numbers
void hex4Print(int temp) {
  if (temp < 0x10) {
    Serial.print("0");
  }
  if (temp < 0x100) {
    Serial.print("0");
  }
  if (temp < 0x1000) {
    Serial.print("0");
  }
  Serial.print(temp, HEX);
}

//probably needs to go
char wait_for_char() {
  int c;
  while ( Serial.available() < 1)
    ; // wait for at least one  character to be available
  c = Serial.read();
  Serial.flush();
  return (char) c;
}

//Toggles for control lines OE CE PP and Vpp
void CE(byte m) { //Chip Enable
  if (m == 1) {
    PORTK = PORTK | B01000000;  //Mega pin A14
  }
  else {
    PORTK = PORTK & B10111111;
  }
}
void OE(byte m) { //Output Enable
  if (m == 1) {
    PORTK = PORTK | B00100000;  //Mega pin A13
  }
  else {
    PORTK = PORTK & B11011111;
  }
}
void PP(byte m) {  //Programming Pulse
  if (m == 1) {
    PORTK = PORTK | B10000000;  //Mega pin A15
  }
  else {
    PORTK = PORTK & B01111111;
  }
}
// Change this to port D7, done
void Vpp(byte m) { //Programmin Voltage Enable
  if (m == 1) {
    digitalWrite(7, 1); //Programming voltage on
  }
  else {
    digitalWrite(7, 0);//Programming voltage off
  }
}

//Strip the control pins from the PORT
byte funcPins() {
  return PORTK & 0xE0; //Upper 3 pins areOE(5), CE(6), PP(7)
}

//Gets a single Uppercase character to be checked for functions
char getChar() {
  for (int j = 0; j < 14; j++) {
    inBuff[j] = 0; //terminators in string format
  }
  while ( Serial.available() < 1);
  Serial.readBytesUntil(13, inBuff, 12);
  if (inBuff[0] > 96) {
    inBuff[0] = inBuff[0] - 32;
  }
  Serial.print(inBuff);
  return (char) inBuff[0];
}

//Allows a Uppercase filename to be set load and store operations
void setFilename() {
  for (int j = 0; j < 14; j++) {
    nomifile[j] = 0; //terminators in string format
  }
  Serial.print("Enter the filename = ");
  while ( Serial.available() < 1);
  Serial.readBytesUntil(13, nomifile, 12);
  for (int j = 0; j < 14; j++) {
    if (nomifile[j] > 96) {
      nomifile[j] = nomifile[j] - 32; //make all characters Upper Case
    }
  }
  Serial.print(nomifile);
  existingfile();
}

// Checks for an existing file name and alerts user
void existingfile() {
  if (SD.exists(nomifile)) {
    Serial.println(" (Exists....)");
  }
  else {
    Serial.println(" (No such file)");
  }
}


//Example shown in SD Library www.ardunio.com.cc
void printDirectory(File dir, int numTabs) {
  while (true) {
    File entry =  dir.openNextFile();
    if (! entry) {
      // no more files
      //Serial.println("**nomorefiles**");
      break;
    }
    for (uint8_t i = 0; i < numTabs; i++) {
      Serial.print('\t');
    }
    Serial.print(entry.name());
    if (entry.isDirectory()) {
      Serial.println("/");
      printDirectory(entry, numTabs + 1);
    }
    else {
      // files have sizes, directories do not
      Serial.print("\t\t");
      Serial.println(entry.size(), DEC);
    }
  }
}
//Crunch an ASCII key to an Integer
unsigned int get_int() {
  int c;
  unsigned int i = 0;
  while (1) {
    while ((c = Serial.read()) <= 0)
      ; // spin waiting for char
    Serial.print(c);
    if (c < '0' || c > '9')
      break;
    i = (i * 10) + (c - '0');
  }
  return i;
}

/* PORT access
  eg
  PORT "D" maps to Arduino digital pins 0 to 7
  DDRD - The Port D Data Direction Register - 0=input/1=output
  PORTD -The Port D Data Register - read/write of Output Register status
  PIND - The Port D Input Pins Register - reflects status of the external pins


  Check ALL THE BELOW FOR NEW WIRING


  http://www.futurlec.com/ICEEPROM.shtml  for EEPROMS

 **EPROM Pin Functions**

  A general observation is that all EPROMs Vpp voltage is either 5V or Vpp
  except 2732 which the Vpp becomes G and must be at 0V with CE to read
  and the 2532 which has Vpp at 5V and CE at 0V CE to read.
  The other pins are logic levels.

  Most have Chip enables that are negative levels
  Some are programming pulse positive eg 2716 pulses EP High, 2708 (pulses Vpp High while CS low)
  Others are programming pulse negative eg 2732, 2532, 2764
  Some can Vpp applied continously 2764, but not 2732


  EPROM 2716, 2k by 8Bits
  Vpp=25V (+0.1uF Cer), 50ms Active low on OE per byte
  Prog Enable  P20=PE , Programming enable = 1, otherwise 0 disables effect of 25V
  Chip Enable  P18=OE , Pulsed high for 50ms, while P10 is at 5V
  Programming: P21=25V, P20=Vpp  P18=5V (Programming True)
  Reading    : P21=5V , P20=0V   P18=0V (Reading True)

  Set up P20&P18 lo, set Address and Data with P20 hi, then pulse P18 hi for 50ms
  Return P20 to lo and data should be readable.

  A7   P1 ----U----P24   +5V
  A6   P2 -       -P23   A8
  A5   P3 -       -P22   A9
  A4   P4 -       -P21   Vpp
  A3   P5 -       -P20   Not(CE), EP
  A2   P6 -       -P19   A10
  A1   P7 -       -P18   Not(OE),
  A0   P8 -       -P17   D7
  D0   P9 -       -P16   D6
  D1   P10-       -P15   D5
  D2   P11-       -P14   D4
  Gnd  P12---------P13   D3

  Read Cycle
  A0-A10 stable
  Vpp 5V
  EP 1 disabled, 0 enabled
  G  1 disabled, 0 enabled
  Program Cycle
  A0-A11, D0-D7 stable
  Vpp 5V to 21V
  EP 0 to  1  to 0 is programming pulse
  G  0 to 111 to 0 enbles programming

  EPROM 2732A, 4k by 8Bits
  Vpp=21V (+0.1uF Cer), 50ms Active low Programming pulse on OE per byte
  Programming: P21=A11, P20=Vpp=21V  P18=0Vpulse (Programming True)
  Reading    : P21=A11, P20=0V       P18=0V (Reading True)

  A7   P1 ----U----P24   +5V
  A6   P2 -       -P23   A8
  A5   P3 -       -P22   A9
  A4   P4 -       -P21   A11
  A3   P5 -       -P20   Not(GVpp)
  A2   P6 -       -P19   A10
  A1   P7 -       -P18   Not(OE)
  A0   P8 -       -P17   D7
  D0   P9 -       -P16   D6
  D1   P10-       -P15   D5
  D2   P11-       -P14   D4
  Gnd  P12---------P13   D3

  Read Cycle
  A0-A11 stable
  Vpp 5V
  OE 1 disabled, 0 enabled
  Program Cycle
  A0-A11, D0-D7 stable
  Vpp 5V to 21V
  OE 1 to 0 to 1 is programming pulse



  EPROM TMS 2532A, 4k by 8Bits
  Vpp=25V (+0.1uF Cer), 50ms Active low Programming pulse on PD per byte
  Programming: P21=Vpp=25V, P20=PGM=0 pulse  P18=A11
  Reading    : P21=5V,      P20=0            P18=A11

  A7   P1 ----U----P24   +5V
  A6   P2 -       -P23   A8
  A5   P3 -       -P22   A9
  A4   P4 -       -P21   Vpp
  A3   P5 -       -P20   PD/Not(PGM)
  A2   P6 -       -P19   A10
  A1   P7 -       -P18   A11
  A0   P8 -       -P17   D7
  D0   P9 -       -P16   D6
  D1   P10-       -P15   D5
  D2   P11-       -P14   D4
  Gnd  P12---------P13   D3


  Read Cycle
  A0-A11 stable
  Vpp 5V
  PD 1 disabled, 0 enabled
  Program Cycle
  A0-A11, D0-D7 stable
  Vpp 5V to 25V
  PD 1 to 0 to 1 is programming pulse


  EPROM 2764, 8k by 8Bits

  Vpp  P1 ----U----P28   +5V
  A12  P2 -       -P27   Not(PGM)
  A7   P3 -       -P26   N/C
  A6   P4 -       -P25   A8
  A5   P5 -       -P24   A9
  A4   P6 -       -P23   A11
  A3   P7 -       -P22   Not(OE) G
  A2   P8 -       -P21   A10
  A1   P9 -       -P20   Not(CE) E
  A0   P10-       -P19   D7
  D0   P11-       -P18   D6
  D1   P12-       -P17   D5
  D2   P13-       -P16   D4
  Gnd  P14---------P15   D3


  Program Cycle
  Vpp = 12.5V (Can be continous).
  PGM is pulse low, and E is low. G is high
  Read Cycle
  PGM is high, E low and G low.
  This chips has several voltage levels  that
  are supposed to be set. However a successful
  hack is to just switch Vpp manually with a
  single pole switch beteen Vcc and Vpp.
  It seems to work ok......

*/
